// File: BaseInterface.h
// Description: Definition of base interface and smart pointer to interface.
// Copyright: (C) 2005 EffectiveSoft Ltd. All Rights Reserved.
// Technical support: technical-support@effective-soft.com

#if !defined(SEM_SDK_STBASE_BASE_INTERFACE_H)
#define SEM_SDK_STBASE_BASE_INTERFACE_H

#if defined(_MSC_VER) && (_MSC_VER >= 1000)
# pragma once
#endif

#include <cstddef>

#if !defined(STBASE_API)
# define STBASE_API
#endif

namespace NsSemSDK
{

/// base class for all SDK exceptions
class STBASE_API CSemBaseException
{
public:
	/// Constructor
	/// @param [in] pszThrowFile: name of source file containing code that throw exception
	/// @param [in] nThrowLine: line number in source file containing code that throw exception
	/// @param [in] nErrorCode: code of general description of error
	CSemBaseException(const char* pszThrowFile = NULL, int nThrowLine = 0, int nErrorCode = 0)
		: m_pszThrowFile(pszThrowFile), m_nThrowLine(nThrowLine), m_nErrorCode(nErrorCode), m_nSystemCode(0)
	{
	}
	/// Destructor
	virtual ~CSemBaseException() {}

	/// @return text of error message
	virtual const char* what() const 
	{
		return "Unknown exception in semantic module";
	}
	/// @return name of source file containing code that throw exception
	virtual const char* GetThrowFile() const
	{
		return m_pszThrowFile;
	}
	/// @return line number in source file containing code that throw exception
	virtual int GetThrowLine() const
	{
		return m_nThrowLine;
	}
	/// @return code of general description of error
	virtual int GetErrorCode() const
	{
		return m_nErrorCode;
	}
	/// @return text of general error message
	virtual const char* GetErrorMessage() const
	{
		return "Unknown error";
	}
	/// @return system error code
	virtual int GetSystemCode() const
	{
		return m_nSystemCode;
	}
	/// @return system error message
	virtual const char* GetSystemMessage() const
	{
		return "System error is not obtained";
	}

protected:
	void SetSystemCode(int nCode)
	{
		m_nSystemCode = nCode;
	}

private:
	/// code of general description of error
	int m_nErrorCode;
	/// name of source file containing code that throw exception
	const char* m_pszThrowFile;
	/// line number in source file containing code that throw exception
	int m_nThrowLine;
	/// system error code
	int m_nSystemCode;
};

/// Base interface.
class IDestructible
{
public:
	/// Destroys this object.
	virtual void Destroy() throw() = 0;
};

/// Base interface with clean function.
class ICleanable : public IDestructible
{
public:
	/// Effect: release all resources used by object
	virtual void Clean() = 0;
};

class CSemString;

/// contain string text
class ISemString : public IDestructible
{
public:
	/// @return length of string 
	virtual int GetLength() const = 0;
	/// @return text of string
	virtual const char* GetText() const = 0;
	/// @return pointer to implementation object
	virtual CSemString* GetString() = 0;
};

/// Destruction strategy for InterfaceT.
/// @param InterfaceT:
///   Interface. Normally InterfaceT should be either IDestructible or
///   interface that extends it.
template<typename InterfaceT>
class CDestructionStrategy
{
public:
	/// Destroys pInterface.
	/// @param [in] pInterface: Pointer to interface or null pointer.
	static void Destroy(InterfaceT* pInterface);
};

template<typename InterfaceT>
inline void CDestructionStrategy<InterfaceT>::Destroy(
	InterfaceT* const pInterface)
{
	if (pInterface != NULL)
	{
		pInterface->Destroy();
	}
}

template<
	typename InterfaceT,
	typename DestructionStrategyT = CDestructionStrategy<InterfaceT> >
class CInterfacePtr;

/// Reference to CInterfacePtr.
template<
	typename InterfaceT,
	typename DestructionStrategyT = CDestructionStrategy<InterfaceT> >
struct CInterfacePtrRef
{
	CInterfacePtrRef(CInterfacePtr<InterfaceT, DestructionStrategyT>& ref) : ref(ref) {}
	CInterfacePtr<InterfaceT, DestructionStrategyT>& ref;
};

/// Smart pointer to interface.
/// @param InterfaceT:
///   Interface. Normally InterfaceT should be either IDestructible or
///   interface that extends it.
/// @param DestructionStrategyT:
///   Destruction strategy for InterfaceT. You may either provide your
///   own strategy or create specialization of CDestructionStrategy.
template<typename InterfaceT, typename DestructionStrategyT>
class CInterfacePtr
{
public:
	/// Constructor.
	/// @param [in] pInterface: Pointer to interface or null pointer.
	CInterfacePtr(InterfaceT* pInterface = NULL);

	/// Copy constructor.
	/// Note, that copy constructor transfers ownership. This behaviour is
	/// similar to the behaviour of std::auto_ptr copy constructor.
	CInterfacePtr(CInterfacePtr& pInterface);

	CInterfacePtr(CInterfacePtrRef<InterfaceT, DestructionStrategyT> ref);

	/// Destructor.
	~CInterfacePtr();

	typedef void (CInterfacePtr::*SafeBool)() const;

	/// Tests for null.
	/// @return
	///   Valid pointer if the pointer to interface is not a null pointer,
	///   null pointer otherwise.
	operator SafeBool() const
	{
		return m_pInterface != NULL ?
			&CInterfacePtr::ThisClassDoesNotSupportComparisons : NULL;
	}

	/// Converts CInterfacePtr to CInterfacePtrRef.
	operator CInterfacePtrRef<InterfaceT, DestructionStrategyT>()
	{
		return CInterfacePtrRef<InterfaceT, DestructionStrategyT>(*this);
	}

	/// Assignment operator.
	/// Note, that assignment operator transfers ownership. This behaviour is
	/// similar to the behaviour of std::auto_ptr assignment operator.
	CInterfacePtr& operator=(CInterfacePtr& pInterface);

	/// Assignment operator.
	/// Note, that assignment operator transfers ownership. This behaviour is
	/// similar to the behaviour of std::auto_ptr assignment operator.
	CInterfacePtr& operator=(CInterfacePtrRef<InterfaceT, DestructionStrategyT> ref);

	/// Returns the stored pointer.
	/// @return The stored pointer to interface.
	InterfaceT* operator->() const
	{
		return m_pInterface;
	}

	/// Dereferences pointer.
	/// @return Reference to object that implements interface.
	InterfaceT& operator*() const
	{
		return *m_pInterface;
	}

	/// Returns the stored pointer.
	/// @return The stored pointer to interface.
	InterfaceT* Get() const
	{
		return m_pInterface;
	}

	/// Resets stored pointer to a new value, destroying object pointed
	/// by the previously stored pointer. Does nothing if pInterface
	/// equals to stored pointer.
	/// @param [in] pInterface: New pointer value.
	void Reset(InterfaceT* pInterface = NULL);

	/// Address-of operator.
	/// Resets the stored pointer and returns its address.
	/// @return Address of the stored pointer.
	InterfaceT** operator&();

	/// Replaces the stored pointer with a null pointer and returns the
	/// previously stored pointer.
	/// @return The previously stored pointer.
	InterfaceT* Detach();

private:
	void ThisClassDoesNotSupportComparisons() const {}

	InterfaceT* m_pInterface; ///< Stored pointer to interface.
};

template<typename InterfaceT, typename DestructionStrategyT>
CInterfacePtr<InterfaceT, DestructionStrategyT>::CInterfacePtr(
	InterfaceT* const pInterface)
:	m_pInterface(pInterface)
{
}

template<typename InterfaceT, typename DestructionStrategyT>
CInterfacePtr<InterfaceT, DestructionStrategyT>::CInterfacePtr(
	CInterfacePtr& pInterface)
:	m_pInterface(pInterface.m_pInterface)
{
	pInterface.m_pInterface = NULL;
}

template<typename InterfaceT, typename DestructionStrategyT>
CInterfacePtr<InterfaceT, DestructionStrategyT>::CInterfacePtr(
	CInterfacePtrRef<InterfaceT, DestructionStrategyT> ref)
:	m_pInterface(ref.ref.m_pInterface)
{
	ref.ref.m_pInterface = NULL;
}

template<typename InterfaceT, typename DestructionStrategyT>
CInterfacePtr<InterfaceT, DestructionStrategyT>::~CInterfacePtr()
{
	DestructionStrategyT::Destroy(m_pInterface);
}

template<typename InterfaceT, typename DestructionStrategyT>
CInterfacePtr<InterfaceT, DestructionStrategyT>&
CInterfacePtr<InterfaceT, DestructionStrategyT>::operator=(
	CInterfacePtr& pInterface)
{
	if (m_pInterface != pInterface.m_pInterface)
	{
		Reset(pInterface.m_pInterface);
		pInterface.m_pInterface = NULL;
	}
	return *this;
}

template<typename InterfaceT, typename DestructionStrategyT>
CInterfacePtr<InterfaceT, DestructionStrategyT>&
CInterfacePtr<InterfaceT, DestructionStrategyT>::operator=(
	CInterfacePtrRef<InterfaceT, DestructionStrategyT> ref)
{
	if (m_pInterface != ref.ref.m_pInterface)
	{
		Reset(ref.ref.m_pInterface);
		ref.ref.m_pInterface = NULL;
	}
	return *this;
}

template<typename InterfaceT, typename DestructionStrategyT>
void CInterfacePtr<InterfaceT, DestructionStrategyT>::Reset(
	InterfaceT* const pInterface)
{
	if (m_pInterface != pInterface)
	{
		DestructionStrategyT::Destroy(m_pInterface);
		m_pInterface = pInterface;
	}
}

template<typename InterfaceT, typename DestructionStrategyT>
InterfaceT** CInterfacePtr<InterfaceT, DestructionStrategyT>::operator&()
{
	Reset();
	return &m_pInterface;
}

template<typename InterfaceT, typename DestructionStrategyT>
InterfaceT* CInterfacePtr<InterfaceT, DestructionStrategyT>::Detach()
{
	InterfaceT* pInterface = m_pInterface;
	m_pInterface = NULL;
	return pInterface;
}

template<
	typename Interface1T, typename DestructionStrategy1T,
	typename Interface2T, typename DestructionStrategy2T>
bool operator==(
	const CInterfacePtr<Interface1T, DestructionStrategy1T>& pLHS,
	const CInterfacePtr<Interface2T, DestructionStrategy2T>& pRHS)
{
	return pLHS.Get() == pRHS.Get();
}

template<
	typename Interface1T, typename DestructionStrategy1T,
	typename Interface2T, typename DestructionStrategy2T>
bool operator!=(
	const CInterfacePtr<Interface1T, DestructionStrategy1T>& pLHS,
	const CInterfacePtr<Interface2T, DestructionStrategy2T>& pRHS)
{
	return pLHS.Get() != pRHS.Get();
}

} // namespace NsSemSDK

#endif // !defined(SEM_SDK_STBASE_BASE_INTERFACE_H)
